#ifndef FILE_SVDTREFFTZ_HPP
#define FILE_SVDTREFFTZ_HPP
#include <comp.hpp>
#include <python_comp.hpp>

#ifndef FILE_INTEGRATORCFHPP
#include <integratorcf.hpp>
#define FILE_INTEGRATORCFHPP
#endif

/// For a mesh element, this struct holds the ndof for the Trefftz and
/// conforming part of the conforming Trefftz space.
template <typename SCAL> struct ElmatWithTrefftzInfo
{
  ngla::Matrix<SCAL> elmat;
  size_t ndof_trefftz;
};

namespace ngcomp
{
  /// creates an embedding marix P for the given operations `op`,
  /// `cop_lhs`, `cop_rhs`.
  ///
  /// The embedding is subject to the constraints specified in
  /// `cop_lhs` and `cop_rhs`.
  ///
  ///  @param op the differential operation. Bilinear from on `fes` x
  ///  `fes_test`.
  ///
  ///  @param fes the trial finite element space of `op`
  ///
  ///  @param fes_test the test finite element space of `op`
  ///
  ///  @param cop_lhs left hand side of the conformity operation. Bilinear form
  ///  on `fes` x `fes_conformity`.
  ///
  ///  @param cop_rhs right hand side of the conformity operation. Bilinear
  ///  form on `fes_conformity` x `fes_conformity`.
  ///
  ///  @param fes_conformity finite element space of the conformity operation
  ///
  ///  @param linear_form right hand sinde to the Trefftz operation `op`.
  ///  Linear form on `fes_test`.
  ///
  ///  @param ndof_trefftz number of degrees of freedom per element
  ///      in the Trefftz finite element space on `fes`, generated by `op`
  ///      (i.e. the local dimension of the kernel of `op` on one element).
  ///      If this is `nullopt`, then `ndof_trefftz` will be set to `ndof -
  ///      ndof_conforming` for each element. Thus it is assumed, that the
  ///      kernel of `op` is sufficiently big.
  ///
  ///  @return (P, f), the embedding `P` and particlar solution `f`. `P` is
  ///  represented as a vector of all element matrices.
  template <typename SCAL>
  pair<vector<optional<ElmatWithTrefftzInfo<SCAL>>>,
       shared_ptr<ngla::BaseVector>>
  EmbTrefftz (const std::optional<SumOfIntegrals> &op, const FESpace &fes,
              const FESpace &fes_test,
              const std::optional<ngfem::SumOfIntegrals> &cop_lhs,
              const std::optional<ngfem::SumOfIntegrals> &cop_rhs,
              const shared_ptr<const FESpace> fes_conformity,
              shared_ptr<const ngfem::SumOfIntegrals> linear_form,
              const std::variant<size_t, double> ndof_trefftz,
              shared_ptr<std::map<std::string, Vector<SCAL>>> stats = nullptr,
              const bool get_range = false);

  /// Represents the FESpace, that is generated by the embedded Trefftz method.
  /// Use the \ref EmbTrefftzFESpace(shared_ptr<T>) constructor to build the
  /// space from a given (trial) FESpcae, use \ref EmbTrefftzFESpace::SetOp to
  /// specify the operation that the Trefftz space should be build from.
  ///
  /// @tparam T must be a FESpace
  template <typename T>
  class EmbTrefftzFESpace
      : public T //, public std::enable_shared_from_this<EmbTrefftzFESpace>
  {
    static_assert (std::is_base_of_v<FESpace, T>, "T must be a FESpace");
    vector<optional<ElmatWithTrefftzInfo<double>>> ETmats;
    vector<optional<ElmatWithTrefftzInfo<Complex>>> ETmatsC;
    shared_ptr<T> fes;
    shared_ptr<FESpace> fes_conformity;
    Array<DofId> all2comp;

  public:
    EmbTrefftzFESpace (shared_ptr<MeshAccess> ama, const Flags &flags,
                       bool parseflags = false)
        : T (ama, flags, parseflags)
    {
      throw Exception ("Please provide a base fes for the embedding");
    }

    EmbTrefftzFESpace (shared_ptr<T> afes)
        : T (afes->GetMeshAccess (), afes->GetFlags (), false), fes (afes)
    {
      this->name = "EmbTrefftzFESpace";
      this->type = "embt";
      this->needs_transform_vec = true;
      this->iscomplex = afes->IsComplex ();
      shared_ptr<CompoundFESpace> cafes
          = dynamic_pointer_cast<CompoundFESpace> (afes);
      if (cafes)
        for (auto space : cafes->Spaces ())
          dynamic_cast<CompoundFESpace *> (this)->AddSpace (space);
      // this->Update();
      // this->UpdateDofTables();
      // this->UpdateCouplingDofArray();
      // this->FinalizeUpdate();
    }

    shared_ptr<BaseVector>
    SetOp (shared_ptr<SumOfIntegrals> bf, shared_ptr<SumOfIntegrals> lf,
           double eps, shared_ptr<FESpace> test_fes, int tndof);

    /// sets up the space for the conforming Trefftz method.
    ///
    ///  @param op the differential operation. Bilinear from on `fes` x
    ///  `fes_test`.
    ///
    ///  @param fes_test the test finite element space of `op`
    ///
    ///  @param cop_lhs left hand side of the conformity operation. Bilinear
    ///  form on `fes` x `fes_conformity`.
    ///
    ///  @param cop_rhs right hand side of the conformity operation. Bilinear
    ///  form on `fes_conformity` x `fes_conformity`.
    ///
    ///  @param fes_conformity finite element space of the conformity operation
    ///
    ///  @param linear_form right hand sinde to the Trefftz operation `op`.
    ///  Linear form on `fes_test`.
    ///
    ///  @param ndof_trefftz number of degrees of freedom per element
    ///      in the Trefftz finite element space on `fes`, generated by `op`
    ///      (i.e. the local dimension of the kernel of `op` on one element)
    ///
    /// @returns the particular solution of the Trefftz setup.
    ///
    /// @throws std::invalid_argument if `op`, `cop_lhs`, or `cop_rhs` are
    /// `null`.
    shared_ptr<BaseVector>
    SetOp (shared_ptr<const SumOfIntegrals> op,
           shared_ptr<const SumOfIntegrals> cop_lhs,
           shared_ptr<const SumOfIntegrals> cop_rhs,
           shared_ptr<const FESpace> fes_conformity,
           shared_ptr<const FESpace> fes_test,
           shared_ptr<const SumOfIntegrals> linear_form, size_t ndof_trefftz);

    void GetDofNrs (ElementId ei, Array<int> &dnums) const override;

    virtual void VTransformMR (ElementId ei, const SliceMatrix<double> mat,
                               TRANSFORM_TYPE type) const override;

    virtual void VTransformMC (ElementId ei, const SliceMatrix<Complex> mat,
                               TRANSFORM_TYPE type) const override;

    virtual void VTransformVR (ElementId ei, const SliceVector<double> vec,
                               TRANSFORM_TYPE type) const override;

    virtual void VTransformVC (ElementId ei, const SliceVector<Complex> vec,
                               TRANSFORM_TYPE type) const override;

    shared_ptr<GridFunction> Embed (shared_ptr<GridFunction> tgfu);

    shared_ptr<BaseMatrix> GetEmbedding ();

  private:
    /// adjusts the dofs of the space. Will be called by SetOp.
    void adjustDofsAfterSetOp ();
  };
}

#ifdef NGS_PYTHON
#include <python_ngstd.hpp>
void ExportEmbTrefftz (py::module m);
#endif // NGS_PYTHON

#endif
